"""
Модуль, содержащий схемы репозиториев.
"""

import uuid
from typing import Generic, TypeVar

from pydantic import BaseModel

IdType = TypeVar('IdType')


class CreateSchemaGeneric(BaseModel, Generic[IdType]):
    """
    Схема для создания модели.
    """

    id: IdType | None = None


class ReadSchemaGeneric(BaseModel, Generic[IdType]):
    """
    Схема для чтения модели.
    """

    id: IdType


class UpdateSchemaGeneric(BaseModel, Generic[IdType]):
    """
    Схема для обновления модели.
    """

    id: IdType


CreateSchemaOld = CreateSchemaGeneric[int]
ReadSchemaOld = ReadSchemaGeneric[int]
UpdateSchemaOld = UpdateSchemaGeneric[int]

CreateSchema = CreateSchemaGeneric[uuid.UUID]
ReadSchema = ReadSchemaGeneric[uuid.UUID]
UpdateSchema = UpdateSchemaGeneric[uuid.UUID]



"""
Пакет, содержащий сервис криптографии для шифрования секретных параметров.
"""

from typing import Protocol, Self

from .aes import AesCbcCryptographyService, AesGcmCryptographyService
from .enums import CryptographicAlgorithmEnum


class CryptographyServiceProtocol(Protocol):
    """
    Протокол сервиса криптографии для шифрования секретных параметров.
    """

    def encrypt(self: Self, data: str) -> str:
        """
        Зашифровываем данные.
        """
        ...

    def decrypt(self: Self, encrypted_data: str) -> str:
        """
        Расшифровываем данные.
        """
        ...


class CryptographyServiceFactoryProtocol(Protocol):
    """
    Протокол фабрики сервисов криптографии для шифрования секретных параметров.
    """

    async def make(self: Self, algorithm: CryptographicAlgorithmEnum) -> CryptographyServiceProtocol:
        """
        Создаем сервис криптографии для шифрования секретных параметров.
        """
        ...


class CryptographyServiceFactoryImpl:
    """
    Реализация фабрики сервисов криптографии для шифрования секретных параметров.
    """

    def __init__(self, secret_key: str) -> None:
        self.secret_key = secret_key

    async def make(self: Self, algorithm: CryptographicAlgorithmEnum) -> CryptographyServiceProtocol:
        """
        Создаем сервис криптографии для шифрования секретных параметров.
        """
        match algorithm:
            case CryptographicAlgorithmEnum.AES_GCM:
                return AesGcmCryptographyService(self.secret_key)
            case CryptographicAlgorithmEnum.AES_CBC:
                return AesCbcCryptographyService(self.secret_key)



"""
Роутер приложения healthcheck.
"""

from fastapi import APIRouter

from fast_clean.schemas import StatusOkResponseSchema

router = APIRouter(prefix='/health', tags=['Healthcheck'], include_in_schema=False)


@router.get('')
async def get_healthcheck_status() -> StatusOkResponseSchema:
    """
    Получаем статус сервера.
    """
    return StatusOkResponseSchema()



"""
Модуль, содержащий вспомогательные функции для работы с Pydantic.
"""

import importlib
from collections.abc import Iterable

from pydantic import BaseModel


def rebuild_schemas(modules: Iterable[str]) -> None:
    """
    Перестраиваем не полностью объявленные схемы из-за циклических зависимостей.

    https://docs.pydantic.dev/2.10/concepts/models/#rebuilding-model-schema
    """
    schemas: dict[str, type[BaseModel]] = {}
    for module in modules:
        for key, obj in importlib.import_module(module).__dict__.items():
            if isinstance(obj, type) and issubclass(obj, BaseModel) and obj is not BaseModel:
                schemas[key] = obj
    for schema in schemas.values():
        schema.model_rebuild(_types_namespace=schemas)



"""
Модуль, содержащий модели.
"""

import datetime as dt

from sqlalchemy import DateTime
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.sql import func


class TimestampMixin:
    """
    Миксин, содержащий дату и время создания и обновления модели.
    """

    created_at: Mapped[dt.datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: dt.datetime.now(dt.UTC),
        server_default=func.now(),
    )
    """
    Дата и время создания.
    """
    updated_at: Mapped[dt.datetime] = mapped_column(
        DateTime(timezone=True),
        default=lambda: dt.datetime.now(dt.UTC),
        server_default=func.now(),
        onupdate=lambda: dt.datetime.now(dt.UTC),
    )
    """
    Дата и время обновления.
    """



"""
Модуль, содержащий команды криптографии для шифрования секретных параметров.
"""

from typing import Annotated

import typer

from fast_clean.depends import get_container
from fast_clean.services import CryptographicAlgorithmEnum, CryptographyServiceFactoryProtocol
from fast_clean.utils import typer_async


@typer_async
async def encrypt(
    data: Annotated[str, typer.Argument(help='Данные для шифровки.')],
    algorithm: Annotated[
        CryptographicAlgorithmEnum, typer.Option(help='Криптографический алгоритм')
    ] = CryptographicAlgorithmEnum.AES_GCM,
) -> None:
    """
    Зашифровываем данные.
    """
    container = get_container()
    cryptography_service_factory: CryptographyServiceFactoryProtocol = await container.get_by_type(
        CryptographyServiceFactoryProtocol
    )
    cryptography_service = await cryptography_service_factory.make(algorithm)
    print(cryptography_service.encrypt(data))


@typer_async
async def decrypt(
    data: Annotated[str, typer.Argument(help='Данные для расшифровки.')],
    algorithm: Annotated[
        CryptographicAlgorithmEnum, typer.Option(help='Криптографический алгоритм')
    ] = CryptographicAlgorithmEnum.AES_GCM,
) -> None:
    """
    Расшифровываем данные.
    """
    container = get_container()
    cryptography_service_factory: CryptographyServiceFactoryProtocol = await container.get_by_type(
        CryptographyServiceFactoryProtocol
    )
    cryptography_service = await cryptography_service_factory.make(algorithm)
    print(cryptography_service.decrypt(data))


def use_cryptography(app: typer.Typer) -> None:
    """
    Регистрируем команды криптографии для шифрования секретных параметров.
    """

    app.command()(encrypt)
    app.command()(decrypt)



"""
Модуль, содержащий middleware.
"""

from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware


def use_middleware(app: FastAPI, cors_origins: list[str]) -> FastAPI:
    """
    Регистрируем middleware.
    """
    app.add_middleware(
        CORSMiddleware,
        allow_origins=cors_origins,
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )
    return app



"""
Модуль, содержащий перечисления сервиса криптографии для шифрования секретных параметров.
"""

from enum import StrEnum, auto


class CryptographicAlgorithmEnum(StrEnum):
    """
    Криптографический алгоритм.
    """

    AES_GCM = auto()
    """
    Алгоритм AES в режиме GCM.
    """
    AES_CBC = auto()
    """
    Алгоритм AES в режиме CBC.
    """



"""
Модуль, содержащий команды загрузки данных из файлов.
"""

from typing import Annotated

import typer

from fast_clean.depends import get_container
from fast_clean.services import SeedServiceProtocol
from fast_clean.utils import typer_async


@typer_async
async def load_seed(
    path: Annotated[str | None, typer.Argument(help='Путь к директории для загрузки данных.')] = None,
) -> None:
    """
    Загружаем данные из файлов.
    """
    async with get_container() as container:
        seed_service: SeedServiceProtocol = await container.get_by_type(SeedServiceProtocol)
        await seed_service.load_data(path)


def use_load_seed(app: typer.Typer) -> None:
    """
    Регистрируем команды загрузки данных из файлов.
    """

    app.command()(load_seed)



"""
Модуль, содержащий сервис распределенной блокировки.
"""

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import AsyncContextManager, Protocol, Self

from fast_clean.exceptions import LockError
from redis import asyncio as aioredis
from redis.exceptions import LockError as RedisLockError


class LockServiceProtocol(Protocol):
    """
    Протокол сервиса распределенной блокировки.
    """

    def lock(
        self: Self,
        name: str,
        *,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
    ) -> AsyncContextManager[None]:
        """
        Осуществляем распределенную блокировку.
        """
        ...


class RedisLockService:
    """
    Сервис распределенной блокировки с помощью Redis.
    """

    def __init__(self, redis: aioredis.Redis) -> None:
        self.redis = redis

    @asynccontextmanager
    async def lock(
        self: Self,
        name: str,
        *,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking_timeout: float | None = None,
    ) -> AsyncIterator[None]:
        """
        Осуществляем распределенную блокировку.
        """
        try:
            async with self.redis.lock(name, timeout=timeout, sleep=sleep, blocking_timeout=blocking_timeout):
                yield
        except RedisLockError as lock_error:
            raise LockError() from lock_error



"""
Модуль для работы с датой и временем.
"""

import datetime as dt


def ts_now() -> float:
    """
    Возвращает текущий timestamp по GMT.

    :return: Значение в секундах
    """
    return dt.datetime.now(dt.timezone.utc).timestamp()



"""
Пакет, содержащий репозиторий файлового хранилища.

Представлено 2 реализации:
- Local
- S3
"""

from pathlib import Path
from typing import AsyncContextManager, Protocol, Self

from .enums import StorageTypeEnum
from .local import LocalStorageRepository
from .reader import StreamReaderProtocol, StreamReadProtocol
from .s3 import S3StorageRepository
from .schemas import (
    LocalStorageParamsSchema,
    S3StorageParamsSchema,
    StorageParamsSchema,
)


class StorageRepositoryProtocol(Protocol):
    """
    Протокол репозитория файлового хранилища.
    """

    async def exists(self: Self, path: str | Path) -> bool:
        """
        Проверяем существует ли файл.
        """
        ...

    async def listdir(self: Self, path: str | Path) -> list[str]:
        """
        Получаем список файлов и директорий в заданной директории.
        """
        ...

    async def is_file(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли файл по пути.
        """
        ...

    async def is_dir(self: Self, path: str | Path) -> bool:
        """
        Проверяем находится ли директория по пути.
        """
        ...

    async def read(self: Self, path: str | Path) -> bytes:
        """
        Читаем содержимое файла.
        """
        ...

    def stream_read(self: Self, path: str | Path) -> AsyncContextManager[StreamReaderProtocol]:
        """
        Читаем содержимое файла в потоковом режиме.
        """
        ...

    async def write(self: Self, path: str | Path, content: str | bytes) -> None:
        """
        Создаем файл или переписываем существующий.
        """
        ...

    async def stream_write(
        self: Self,
        path: str | Path,
        stream: StreamReadProtocol,
        length: int = -1,
        part_size: int = 0,
    ) -> None:
        """
        Создаем файл или переписываем существующий в потоковом режиме.
        """
        ...

    async def delete(self: Self, path: str | Path) -> None:
        """
        Удаляем файл.
        """
        ...


class StorageRepositoryFactoryProtocol(Protocol):
    """
    Протокол фабрики репозиториев файлового хранилища.
    """

    async def make(self, storage_type: StorageTypeEnum, params: StorageParamsSchema) -> StorageRepositoryProtocol:
        """
        Создаем репозиторий файлового хранилища.
        """
        ...


class StorageRepositoryFactoryImpl:
    """
    Реализация фабрики репозиториев файлового хранилища.
    """

    async def make(self: Self, storage_type: StorageTypeEnum, params: StorageParamsSchema) -> StorageRepositoryProtocol:
        """
        Создаем репозиторий файлового хранилища.
        """
        if storage_type == StorageTypeEnum.S3 and isinstance(params, S3StorageParamsSchema):
            return S3StorageRepository(params)
        elif storage_type == StorageTypeEnum.LOCAL and isinstance(params, LocalStorageParamsSchema):
            return LocalStorageRepository(params)
        raise NotImplementedError()



"""
Модуль, содержащий перечисления файлового хранилища.
"""

from enum import StrEnum, auto


class StorageTypeEnum(StrEnum):
    """
    Тип хранилища.
    """

    S3 = auto()
    """
    Хранилище S3.
    """
    LOCAL = auto()
    """
    Локальное файловое хранилище.
    """



"""
Пакет, содержащий репозитории.
"""

from .cache import CacheManager as CacheManager
from .cache import CacheRepositoryProtocol as CacheRepositoryProtocol
from .cache import InMemoryCacheRepository as InMemoryCacheRepository
from .cache import RedisCacheRepository as RedisCacheRepository
from .crud import CrudRepositoryOldProtocol as CrudRepositoryOldProtocol
from .crud import CrudRepositoryProtocol as CrudRepositoryProtocol
from .crud import DbCrudRepository as DbCrudRepository
from .crud import DbCrudRepositoryOld as DbCrudRepositoryOld
from .settings import EnvSettingsRepository as EnvSettingsRepository
from .settings import SettingsRepositoryError as SettingsRepositoryError
from .settings import SettingsRepositoryFactoryImpl as SettingsRepositoryFactoryImpl
from .settings import SettingsRepositoryFactoryProtocol as SettingsRepositoryFactoryProtocol
from .settings import SettingsRepositoryProtocol as SettingsRepositoryProtocol
from .settings import SettingsSourceEnum as SettingsSourceEnum
from .storage import LocalStorageParamsSchema as LocalStorageParamsSchema
from .storage import LocalStorageRepository as LocalStorageRepository
from .storage import S3StorageParamsSchema as S3StorageParamsSchema
from .storage import S3StorageRepository as S3StorageRepository
from .storage import StorageRepositoryFactoryImpl as StorageRepositoryFactoryImpl
from .storage import (
    StorageRepositoryFactoryProtocol as StorageRepositoryFactoryProtocol,
)
from .storage import StorageRepositoryProtocol as StorageRepositoryProtocol
from .storage import StorageTypeEnum as StorageTypeEnum
from .storage import StreamReaderProtocol as StreamReaderProtocol
from .storage import StreamReadProtocol as StreamReadProtocol



"""
Модуль, содержащий переменные типов.
"""

import uuid
from typing import TypeVar

from fast_clean.db import Base, BaseOld
from fast_clean.schemas import (
    CreateSchema,
    CreateSchemaOld,
    ReadSchema,
    ReadSchemaOld,
    UpdateSchema,
    UpdateSchemaOld,
)

ModelBaseType = TypeVar('ModelBaseType', bound=BaseOld | Base)
CreateSchemaBaseType = TypeVar('CreateSchemaBaseType', bound=CreateSchemaOld | CreateSchema)
ReadSchemaBaseType = TypeVar('ReadSchemaBaseType', bound=ReadSchemaOld | ReadSchema)
UpdateSchemaBaseType = TypeVar('UpdateSchemaBaseType', bound=UpdateSchemaOld | UpdateSchema)
IdType = TypeVar('IdType', bound=int | uuid.UUID)
IdTypeContravariant = TypeVar('IdTypeContravariant', bound=int | uuid.UUID, contravariant=True)


ModelOldType = TypeVar('ModelOldType', bound=BaseOld)
CreateSchemaOldType = TypeVar('CreateSchemaOldType', bound=CreateSchemaOld)
ReadSchemaOldType = TypeVar('ReadSchemaOldType', bound=ReadSchemaOld)
UpdateSchemaOldType = TypeVar('UpdateSchemaOldType', bound=UpdateSchemaOld)


ModelType = TypeVar('ModelType', bound=Base)
CreateSchemaType = TypeVar('CreateSchemaType', bound=CreateSchema)
ReadSchemaType = TypeVar('ReadSchemaType', bound=ReadSchema)
UpdateSchemaType = TypeVar('UpdateSchemaType', bound=UpdateSchema)



"""
Модуль, содержащий тесты контейнера зависимостей.
"""

from typing import cast

import pytest
from fast_clean.container import ContainerImpl, ContainerProtocol
from fast_clean.exceptions import ContainerError

from .depends import (
    RepositoryAImpl,
    RepositoryAProtocol,
    RepositoryBImpl,
    RepositoryBProtocol,
    RepositoryUnknownProtocol,
    ServiceAImpl,
    ServiceAProtocol,
    ServiceBImpl,
    Session,
    UseCaseAImpl,
)


class TestContainer:
    """
    Тесты контейнера зависимостей.
    """

    @staticmethod
    async def test_get_by_name_without_dependencies(
        container: ContainerProtocol,
    ) -> None:
        """
        Тестируем метод `get_by_name` для зависимости, которая ни от чего не зависит.
        """
        repository_a = await container.get_by_name('repository_a')
        assert isinstance(repository_a, RepositoryAImpl)

    @staticmethod
    async def test_get_by_name_unknown(container: ContainerProtocol) -> None:
        """
        Тестируем метод `get_by_name` для неизвестной зависимости.
        """
        with pytest.raises(ContainerError):
            await container.get_by_name('unknown_repository')

    @staticmethod
    async def test_get_by_type_without_dependencies(
        container: ContainerProtocol,
    ) -> None:
        """
        Тестируем метод `get_by_type` для зависимости, которая ни от чего не зависит.
        """
        repository_a = await container.get_by_type(RepositoryAProtocol)
        assert isinstance(repository_a, RepositoryAImpl)

    @staticmethod
    async def test_get_by_type_unknown(container: ContainerProtocol) -> None:
        """
        Тестируем метод `get_by_type` для неизвестной зависимости.
        """
        with pytest.raises(ContainerError):
            await container.get_by_type(RepositoryUnknownProtocol)

    @staticmethod
    async def test_get_by_name_with_dependencies(container: ContainerImpl) -> None:
        """
        Тестируем метод `get_by_name` для зависимости, которая зависит от других зависимостей.
        """
        with pytest.raises(ContainerError):
            await container.get_by_name('service_a')
        async with container:
            service_a = await container.get_by_name('service_a')
            assert isinstance(service_a, ServiceAImpl)
            assert ServiceAProtocol in container.instances
            assert isinstance(service_a.repository_a, RepositoryAImpl)
            assert RepositoryAProtocol not in container.instances
            assert isinstance(service_a.repository_b, RepositoryBImpl)
            assert RepositoryBProtocol in container.instances

    @staticmethod
    async def test_get_by_name_with_extra(container: ContainerProtocol) -> None:
        """
        Тестируем метод `get_by_name` для большого дерева зависимостей, содержащего дополнительные параметры.
        """
        expected_int_value = 10
        expected_str_value = 'str_value'
        session = Session()
        async with container:
            use_case_a = await container.get_by_name(
                'use_case_a',
                extra={
                    ('session', Session | None): session,
                    ('value', int): lambda: expected_int_value,
                    ('value', str): expected_str_value,
                },
            )
            assert isinstance(use_case_a, UseCaseAImpl)
            assert isinstance(use_case_a.service_a, ServiceAImpl)
            assert isinstance(use_case_a.service_a.repository_a, RepositoryAImpl)
            assert isinstance(use_case_a.service_a.repository_b, RepositoryBImpl)
            assert isinstance(use_case_a.service_b, ServiceBImpl)
            assert isinstance(use_case_a.service_b.repository_a, RepositoryAImpl)
            assert isinstance(use_case_a.service_b.repository_b, RepositoryBImpl)
            assert use_case_a.service_b.value == expected_int_value
            assert use_case_a.value == expected_str_value
            assert use_case_a.service_a.repository_a is not use_case_a.service_b.repository_a
            assert use_case_a.service_a.repository_b is use_case_a.service_b.repository_b
            assert session.in_transaction
        assert not session.in_transaction

    @staticmethod
    async def test_create_new_with_call(container: ContainerImpl) -> None:
        """
        Тестируем создание нового контейнера с помощью вызова.
        """
        expected_int_value = 10
        async with container({('value', int): expected_int_value}) as new_container:
            new_container = cast(ContainerImpl, new_container)
            assert container != new_container
            service_b = await new_container.get_by_name('service_b')
            assert isinstance(service_b, ServiceBImpl)
            assert service_b.value == expected_int_value
            assert new_container.instances
            assert not container.instances



"""
Модуль, содержащий схемы ответов статусов.
"""

from .request_response import ResponseSchema


class StatusOkResponseSchema(ResponseSchema):
    """
    Схема успешного ответа.
    """

    status: str = 'ok'



"""
Модуль, содержащий зависимости.
"""

import json
from collections.abc import AsyncIterator, Sequence
from typing import Annotated

from fastapi import Depends, Request
from faststream.kafka import KafkaBroker
from flatten_dict import unflatten
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.datastructures import FormData
from stringcase import snakecase

from .broker import BrokerFactory
from .container import ContainerImpl, ContainerProtocol
from .db import SessionFactory, SessionManagerImpl, SessionManagerProtocol
from .redis import RedisManager
from .repositories import (
    CacheManager,
    CacheRepositoryProtocol,
    LocalStorageParamsSchema,
    S3StorageParamsSchema,
    SettingsRepositoryFactoryImpl,
    SettingsRepositoryFactoryProtocol,
    SettingsRepositoryProtocol,
    SettingsSourceEnum,
    StorageRepositoryFactoryImpl,
    StorageRepositoryFactoryProtocol,
    StorageRepositoryProtocol,
    StorageTypeEnum,
)
from .schemas import PaginationRequestSchema
from .services import (
    CryptographicAlgorithmEnum,
    CryptographyServiceFactoryImpl,
    CryptographyServiceFactoryProtocol,
    CryptographyServiceProtocol,
    LockServiceProtocol,
    RedisLockService,
    SeedServiceImpl,
    SeedServiceProtocol,
    TransactionServiceImpl,
    TransactionServiceProtocol,
)
from .settings import CoreCacheSettingsSchema, CoreKafkaSettingsSchema, CoreSettingsSchema, CoreStorageSettingsSchema

# --- utils ---


def get_container() -> ContainerProtocol:
    """
    Получаем контейнер зависимостей.
    """
    ContainerImpl.init()
    return ContainerImpl()


async def get_nested_form_data(request: Request) -> FormData:
    """
    Получаем форму, позволяющую использовать вложенные словари.
    """
    dot_data = {k.replace('[', '.').replace(']', ''): v for k, v in (await request.form()).items()}
    nested_data = unflatten(dot_data, 'dot')
    for k, v in nested_data.items():
        if isinstance(v, dict):
            nested_data[k] = json.dumps(v)
    return FormData(nested_data)


def get_pagination(page: int | None = None, page_size: int | None = None) -> PaginationRequestSchema:
    """
    Получаем входные данные пагинации.
    """
    return PaginationRequestSchema(page=page or 1, page_size=page_size or 10)


def get_sorting(sorting: str | None = None) -> Sequence[str]:
    """
    Получаем входные данные сортировки.
    """
    if not sorting:
        return []
    return [s[0] + snakecase(s[1:]) if s[0] == '-' else snakecase(s) for s in sorting.split(',')]


Container = Annotated[ContainerProtocol, Depends(get_container)]
NestedFormData = Annotated[FormData, Depends(get_nested_form_data)]
Pagination = Annotated[PaginationRequestSchema, Depends(get_pagination)]
Sorting = Annotated[Sequence[str], Depends(get_sorting)]

# --- repositories ---


def get_settings_repository_factory() -> SettingsRepositoryFactoryProtocol:
    """
    Получаем фабрику репозиториев настроек.
    """
    return SettingsRepositoryFactoryImpl()


SettingsRepositoryFactory = Annotated[SettingsRepositoryFactoryProtocol, Depends(get_settings_repository_factory)]


async def get_settings_repository(settings_repository_factory: SettingsRepositoryFactory) -> SettingsRepositoryProtocol:
    """
    Получаем репозиторий настроек.
    """
    return await settings_repository_factory.make(SettingsSourceEnum.ENV)


SettingsRepository = Annotated[SettingsRepositoryProtocol, Depends(get_settings_repository)]


async def get_settings(settings_repository: SettingsRepository) -> CoreSettingsSchema:
    """
    Получаем настройки.
    """
    return await settings_repository.get(CoreSettingsSchema)


Settings = Annotated[CoreSettingsSchema, Depends(get_settings)]


async def get_broker_repository(settings_repository: SettingsRepository) -> AsyncIterator[KafkaBroker]:
    """
    Получаем репозиторий брокера сообщений.
    """
    kafka_settings = await settings_repository.get(CoreKafkaSettingsSchema)
    yield BrokerFactory.make_static(kafka_settings)


async def get_cache_repository(settings_repository: SettingsRepository) -> CacheRepositoryProtocol:
    """
    Получаем репозиторий кеша.
    """
    settings = await settings_repository.get(CoreSettingsSchema)
    if settings.redis_dsn is not None:
        RedisManager.init(settings.redis_dsn)
    cache_settings = await settings_repository.get(CoreCacheSettingsSchema)
    if CacheManager.cache is None:
        CacheManager.init(cache_settings, RedisManager.redis)
    if CacheManager.cache is not None:
        return CacheManager.cache
    raise ValueError('Cache is not initialized')


def get_storage_repository_factory() -> StorageRepositoryFactoryProtocol:
    """
    Получаем фабрику репозиториев файлового хранилища.
    """
    return StorageRepositoryFactoryImpl()


BrokerRepository = Annotated[KafkaBroker, Depends(get_broker_repository)]
CacheRepository = Annotated[CacheRepositoryProtocol, Depends(get_cache_repository)]
StorageRepositoryFactory = Annotated[StorageRepositoryFactoryProtocol, Depends(get_storage_repository_factory)]


async def get_storage_repository(
    settings_repository: SettingsRepository,
    storage_repository_factory: StorageRepositoryFactory,
) -> StorageRepositoryProtocol:
    """
    Получаем репозиторий файлового хранилища.
    """
    storage_settings = await settings_repository.get(CoreStorageSettingsSchema)
    if storage_settings.provider == 's3' and storage_settings.s3 is not None:
        return await storage_repository_factory.make(
            StorageTypeEnum.S3,
            S3StorageParamsSchema.model_validate(storage_settings.s3.model_dump()),
        )
    elif storage_settings.provider == 'local':
        return await storage_repository_factory.make(
            StorageTypeEnum.LOCAL, LocalStorageParamsSchema(path=storage_settings.dir)
        )
    raise NotImplementedError(f'Storage {storage_settings.provider} not allowed')


StorageRepository = Annotated[StorageRepositoryProtocol, Depends(get_storage_repository)]

# --- db ---


async def get_async_session(settings_repository: SettingsRepository) -> AsyncIterator[AsyncSession]:
    """
    Получаем асинхронную сессию.
    """
    async with SessionFactory.make_async_session_static(settings_repository) as session:
        yield session


Session = Annotated[AsyncSession, Depends(get_async_session)]


def get_session_manager(session: Session) -> SessionManagerProtocol:
    """
    Получаем менеджер сессий.
    """
    return SessionManagerImpl(session)


SessionManager = Annotated[SessionManagerProtocol, Depends(get_session_manager)]

# --- services ---


def get_cryptography_service_factory(settings: Settings) -> CryptographyServiceFactoryProtocol:
    """
    Получаем фабрику сервисов криптографии.
    """
    return CryptographyServiceFactoryImpl(settings.secret_key)


CryptographyServiceFactory = Annotated[CryptographyServiceFactoryProtocol, Depends(get_cryptography_service_factory)]


async def get_cryptography_service(
    cryptography_service_factory: CryptographyServiceFactory,
) -> CryptographyServiceProtocol:
    """
    Получаем сервис криптографии.
    """
    return await cryptography_service_factory.make(CryptographicAlgorithmEnum.AES_GCM)


def get_lock_service(settings: Settings) -> LockServiceProtocol:
    """
    Получаем сервис распределенной блокировки.
    """
    assert settings.redis_dsn is not None
    RedisManager.init(settings.redis_dsn)
    assert RedisManager.redis is not None
    return RedisLockService(RedisManager.redis)


def get_seed_service(session_manager: SessionManager) -> SeedServiceProtocol:
    """
    Получаем сервис для загрузки данных из файлов.
    """
    return SeedServiceImpl(session_manager)


def get_transaction_service(session: Session) -> TransactionServiceProtocol:
    """
    Получаем сервис транзакций.
    """
    return TransactionServiceImpl(session)


CryptographyService = Annotated[CryptographyServiceProtocol, Depends(get_cryptography_service)]
LockService = Annotated[LockServiceProtocol, Depends(get_lock_service)]
SeedService = Annotated[SeedServiceProtocol, Depends(get_seed_service)]
TransactionService = Annotated[TransactionServiceProtocol, Depends(get_transaction_service)]



"""
Модуль, содержащий функционал, связанный с Redis.
"""

from pydantic import RedisDsn

from redis import asyncio as aioredis


class RedisManager:
    """
    Менеджер для управления клиентом Redis.
    """

    redis: aioredis.Redis | None = None

    @classmethod
    def init(cls, redis_dsn: RedisDsn) -> None:
        """
        Инициализируем клиент Redis.
        """
        if cls.redis is None:
            cls.redis = aioredis.from_url(url=str(redis_dsn), decode_responses=True)



"""
Модуль, содержащий схемы исключений.
"""

from pydantic import BaseModel


class BusinessLogicExceptionSchema(BaseModel):
    """
    Схема базового исключения бизнес-логики.
    """

    type: str
    msg: str
    traceback: str | None


class ModelAlreadyExistsErrorSchema(BusinessLogicExceptionSchema):
    """
    Схема ошибки, возникающей при попытке создать модель с существующим уникальным
    полем.
    """

    field: str


class ValidationErrorSchema(BusinessLogicExceptionSchema):
    """
    Схема ошибки валидации.
    """

    fields: list[str]



"""
Модуль, содержащий контейнер зависимостей.
"""

from __future__ import annotations

import importlib
import inspect
import os
import sys
from collections.abc import Callable
from contextlib import AsyncExitStack
from pathlib import Path
from types import TracebackType
from typing import (
    Annotated,
    Any,
    Protocol,
    Self,
    get_args,
    get_origin,
    get_type_hints,
)

from fastapi.dependencies.utils import (
    is_async_gen_callable,
    is_coroutine_callable,
    is_gen_callable,
    solve_generator,
)
from fastapi.params import Depends
from starlette.concurrency import run_in_threadpool
from stringcase import pascalcase

from .exceptions import ContainerError


class ContainerProtocol(Protocol):
    """
    Протокол контейнера зависимостей.
    """

    def __call__(self: Self, extra: dict[tuple[str, Any], Any] | None = None) -> ContainerProtocol:
        """
        Получаем новый контейнер.
        """
        ...

    async def __aenter__(self: Self) -> ContainerProtocol:
        """
        Открываем асинхронный контекстный менеджер для управления стеком контекстных менеджеров.
        """
        ...

    async def __aexit__(
        self: Self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """
        Закрываем асинхронный контекстный менеджер для управления стеком контекстных менеджеров.
        """
        ...

    async def get_by_type(self: Self, dependency_type: Any, *, extra: dict[tuple[str, Any], Any] | None = None) -> Any:
        """
        Получаем зависимость по типу.
        """
        ...

    async def get_by_name(self: Self, name: str, *, extra: dict[tuple[str, Any], Any] | None = None) -> Any:
        """
        Получаем зависимость по имени.
        """
        ...


class ContainerImpl:
    """
    Реализация контейнера зависимостей.
    """

    DEPENDS_MODULE = 'depends'

    dependencies_name_mapping: dict[str, Annotated[Any, Depends]] | None = None
    dependencies_type_mapping: dict[type, Annotated[Any, Depends]] | None = None

    def __init__(self, extra: dict[tuple[str, Any], Any] | None = None) -> None:
        self.extra = extra
        self.instances: dict[type, Any] = {}
        self.async_exit_stack: AsyncExitStack | None = None

    def __call__(self: Self, extra: dict[tuple[str, Any], Any] | None = None) -> ContainerProtocol:
        """
        Получаем новый контейнер.
        """
        return ContainerImpl(extra)

    async def __aenter__(self: Self) -> ContainerProtocol:
        """
        Открываем асинхронный контекстный менеджер для управления стеком контекстных менеджеров.
        """
        self.async_exit_stack = AsyncExitStack()
        return self

    async def __aexit__(
        self: Self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """
        Закрываем асинхронный контекстный менеджер для управления стеком контекстных менеджеров.
        """
        assert self.async_exit_stack is not None
        await self.async_exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def get_by_type(self: Self, dependency_type: Any, *, extra: dict[tuple[str, Any], Any] | None = None) -> Any:
        """
        Получаем зависимость по типу.
        """
        extra = extra or self.extra or {}
        assert self.dependencies_type_mapping is not None
        if dependency_type not in self.dependencies_type_mapping:
            raise ContainerError(f'Dependency {dependency_type} not found')
        return await self.resolve_dependency(self.dependencies_type_mapping[dependency_type], extra)

    async def get_by_name(self: Self, name: str, *, extra: dict[tuple[str, Any], Any] | None = None) -> Any:
        """
        Получаем зависимость по имени.
        """
        extra = extra or self.extra or {}
        assert self.dependencies_name_mapping is not None
        name = pascalcase(name)
        if name not in self.dependencies_name_mapping:
            raise ContainerError(f'Dependency {name} not found')
        return await self.resolve_dependency(self.dependencies_name_mapping[name], extra)

    async def resolve_dependency(self: Self, annotated_type: type, extra: dict[tuple[str, Any], Any]) -> Any:
        """
        Разрешаем зависимость.
        """
        dependency_type: type
        depends: Depends
        dependency_type, depends = get_args(annotated_type)
        if dependency_type in self.instances:
            return self.instances[dependency_type]
        call = depends.dependency
        assert call is not None
        resolved_params = await self.resolve_params(call, extra)
        if is_gen_callable(call) or is_async_gen_callable(call):
            if self.async_exit_stack is None:
                raise ContainerError('Generator requires `async_exit_stack`')
            resolved = await solve_generator(call=call, stack=self.async_exit_stack, sub_values=resolved_params)
        elif is_coroutine_callable(call):
            resolved = await call(**resolved_params)
        else:
            resolved = await run_in_threadpool(call, **resolved_params)
        if depends.use_cache:
            self.instances[dependency_type] = resolved
        return resolved

    async def resolve_params(self: Self, call: Callable[..., Any], extra: dict[tuple[str, Any], Any]) -> dict[str, Any]:
        """
        Разрешаем параметры зависимости.
        """
        signature = inspect.signature(call)
        resolved_params: dict[str, Any] = {}
        type_hints = get_type_hints(call, include_extras=True)
        for param_name, param in signature.parameters.items():
            annotation = type_hints[param_name]
            args = get_args(annotation)
            if (param_name, annotation) in extra:
                extra_value = extra[param_name, annotation]
                resolved_params[param_name] = extra_value() if callable(extra_value) else extra_value
            elif get_origin(annotation) is Annotated and len(args) == 2 and isinstance(args[1], Depends):
                resolved_params[param_name] = await self.resolve_dependency(annotation, extra)
            elif param.default is not inspect.Parameter.empty:
                resolved_params[param_name] = param.default
            else:
                raise ContainerError(f'Can not resolve dependency {param_name}: {annotation}')
        return resolved_params

    @classmethod
    def init(cls, module_names: set[str] | None = None) -> None:
        """
        Инициализируем контейнер.
        """
        if cls.dependencies_name_mapping is None:
            module_names = module_names or cls.get_default_module_names()
            cls.dependencies_name_mapping = cls.get_dependencies_name_mapping(module_names)
        if cls.dependencies_type_mapping is None:
            cls.dependencies_type_mapping = cls.get_dependencies_type_mapping()

    @classmethod
    def get_default_module_names(cls) -> set[str]:
        """
        Получаем список модулей с зависимостями по умолчанию.
        """
        cwd = Path(os.getcwd())
        virtual_env_paths = {path.parent for path in cwd.rglob('pyvenv.cfg')}
        module_names: set[str] = set()
        for path in cwd.rglob(f'{cls.DEPENDS_MODULE}.py'):
            if not any(path.is_relative_to(venv) for venv in virtual_env_paths):
                module_names.add('.'.join(str(path.relative_to(cwd).with_suffix('')).split('/')))
        module_names.add(f'fast_clean.{cls.DEPENDS_MODULE}')
        return module_names

    @staticmethod
    def get_dependencies_name_mapping(
        module_names: set[str],
    ) -> dict[str, Annotated[Any, Depends]]:
        """
        Получаем маппинг имен на зависимости.
        """
        name_mapping: dict[str, Annotated[Any, Depends]] = {}
        for module_name in module_names:
            module = sys.modules[module_name] if module_name in sys.modules else importlib.import_module(module_name)
            for name, obj in module.__dict__.items():
                args = get_args(obj)
                if get_origin(obj) is Annotated and len(args) == 2 and isinstance(args[1], Depends):
                    name_mapping[name] = obj
        return name_mapping

    @classmethod
    def get_dependencies_type_mapping(cls) -> dict[type, Annotated[Any, Depends]]:
        """
        Получаем маппинг типов на зависимости.
        """
        assert cls.dependencies_name_mapping is not None
        type_mapping: dict[type, Annotated[Any, Depends]] = {}
        for dependency in cls.dependencies_name_mapping.values():
            type_mapping[get_args(dependency)[0]] = dependency
        return type_mapping



"""
Пакет, содержащий сервисы.
"""

from .cryptography import AesCbcCryptographyService as AesCbcCryptographyService
from .cryptography import AesGcmCryptographyService as AesGcmCryptographyService
from .cryptography import CryptographicAlgorithmEnum as CryptographicAlgorithmEnum
from .cryptography import CryptographyServiceFactoryImpl as CryptographyServiceFactoryImpl
from .cryptography import CryptographyServiceFactoryProtocol as CryptographyServiceFactoryProtocol
from .cryptography import CryptographyServiceProtocol as CryptographyServiceProtocol
from .lock import LockServiceProtocol as LockServiceProtocol
from .lock import RedisLockService as RedisLockService
from .seed import SeedServiceImpl as SeedServiceImpl
from .seed import SeedServiceProtocol as SeedServiceProtocol
from .transaction import TransactionServiceImpl as TransactionServiceImpl
from .transaction import TransactionServiceProtocol as TransactionServiceProtocol



"""
Модуль, содержащий тесты функционала, связанного с базой данных.
"""

from typing import cast
from unittest.mock import MagicMock

import sqlalchemy as sa
from fast_clean.db import SessionManagerImpl
from pytest_mock import MockerFixture


class TestSessionManager:
    """
    Тесты менеджера сессий.
    """

    @staticmethod
    async def test_get_session_begin(session_manager: SessionManagerImpl) -> None:
        """
        Тестируем запуск транзакции в методе `begin`.
        """
        assert not session_manager.session.in_transaction()
        async with session_manager.get_session():
            assert session_manager.session.in_transaction()
        assert not session_manager.session.in_transaction()

    @staticmethod
    async def test_get_session_immediate(session_manager: SessionManagerImpl, mocker: MockerFixture) -> None:
        """
        Тестируем запуск транзакции в методе `begin` с параметром `immediate=True`.
        """
        mocker.patch.object(session_manager.session, 'execute')
        assert not session_manager.session.in_transaction()
        async with session_manager.get_session():
            assert session_manager.session.in_transaction()
        execute = cast(MagicMock, session_manager.session.execute)
        mocker.patch.object(sa.text, '__eq__', lambda self, other: str(self) == str(other))
        execute.assert_called_once()
        call_args = execute.call_args[0]
        assert len(call_args) == 1
        assert str(call_args[0]) == str(sa.text('SET CONSTRAINTS ALL IMMEDIATE'))
        assert not session_manager.session.in_transaction()

    @staticmethod
    async def test_get_session_begin_nested(
        session_manager: SessionManagerImpl,
    ) -> None:
        """
        Тестируем запуск дочерней транзакции в методе `begin`.
        """
        async with session_manager.session.begin():
            assert not session_manager.session.in_nested_transaction()
            async with session_manager.get_session():
                assert session_manager.session.in_nested_transaction()
            assert not session_manager.session.in_nested_transaction()



"""
Модуль, содержащий репозиторий настроек, получаемых из переменных окружения.
"""

from typing import Self

from fast_clean.settings import BaseSettingsSchema, CoreSettingsSchema

from .exceptions import SettingsRepositoryError
from .type_vars import SettingsSchema


class EnvSettingsRepository:
    """
    Репозиторий настроек, получаемых из переменных окружения.
    """

    SETTINGS_MODULE = 'settings'

    settings: list[BaseSettingsSchema] | None = None

    async def get(self: Self, schema_type: type[SettingsSchema], *, name: str | None = None) -> SettingsSchema:
        """
        Получаем настройки из переменных окружения.
        """
        if self.settings is None:
            self.settings = [st() for st in BaseSettingsSchema.descendant_types if st is not CoreSettingsSchema]
        if name is not None:
            return self.get_by_name(schema_type, name)
        return self.get_by_type(schema_type)

    def get_by_name(self: Self, schema_type: type[SettingsSchema], name: str) -> SettingsSchema:
        """
        Получаем настройки по имени.
        """
        assert self.settings
        for settings in self.settings:
            value = getattr(settings, name, None)
            if value is not None and isinstance(value, schema_type):
                return value
        raise SettingsRepositoryError(f'Settings with name {name} not found')

    def get_by_type(self: Self, schema_type: type[SettingsSchema]) -> SettingsSchema:
        """
        Получаем настройки по типу.
        """
        assert self.settings
        for settings in self.settings:
            if isinstance(settings, schema_type):
                return settings
            for key in settings.__dict__.keys():
                value = getattr(settings, key, None)
                if isinstance(value, schema_type):
                    return value
        raise SettingsRepositoryError(f'Settings with type {schema_type} not found')



"""
Модуль, содержащий варианты использования.
"""

from __future__ import annotations

from types import new_class
from typing import Any, Protocol, TypeVar, cast

UseCaseResultSchemaType = TypeVar('UseCaseResultSchemaType', covariant=True)


class UseCaseProtocol(Protocol[UseCaseResultSchemaType]):
    """
    Протокол варианта использования.
    """

    async def __call__(self) -> UseCaseResultSchemaType:
        """
        Вызываем вариант использования.
        """
        ...

    def __class_getitem__(cls, params: type | tuple[type, ...]) -> type:
        """
        Создаем уникальный класс.

        По умолчанию одинаковые обобщенные классы указывают на один и тот же _GenericAlias.
        Из-за данного поведения поиск по типу становится невозможным.

            UseCaseA = UseCaseProtocol[None]
            UseCaseB = UseCaseProtocol[None]
            assert UseCaseA is UseCaseB

        Поэтому вместо _GenericAlias возвращаем уникального наследника.
        """
        generic_alias = cast(Any, super()).__class_getitem__(params)
        return new_class(cls.__name__, (generic_alias,))
